var _undoManager = (function (rowCount, colCount) {
    
	var _undoBuffer = [];
	var _redoBuffer = [];
	
	var init = function() {
		_undoBuffer = [];
		_redoBuffer = [];
	};
	
	var addAction = function(action) {
		_redoBuffer = [];
		_undoBuffer.push(action);
	};

	var undoLast = function() {
		var lastAction = _undoBuffer.pop();
		if (lastAction)
		{
			//alert("undoing");
			lastAction.undo();
			_redoBuffer.unshift(lastAction);
		}
	};

	var redoLast = function() {
		var lastAction = _redoBuffer.shift();
		if (lastAction)
		{
			lastAction.redo();
			_undoBuffer.push(lastAction);
		}
	};

		return {
			init : init,
			addAction:addAction,
			undoLast:undoLast,
			redoLast:redoLast
			};
	
})();

jQuery.fn.reverse = [].reverse;

var _aggregateAction = function() {
	
	var _actionBuffer = [];

	var addAction = function(action) {
		_actionBuffer.push(action);
	};

	var undo = function() {
		$(_actionBuffer).reverse().each(function( index, value )	{ value.undo() }	);
	};		

	var redo = function() {
		$.each(_actionBuffer, function( index, value )	{ value.redo() }	);
	};		
	
	return {
		addAction:addAction,
		undo:undo,
		redo:redo
	};
};

var _aggregateActionMgr = function() {
	
	var _curAction = null;

	var init = function() {
		_curAction = new _aggregateAction();
	};

	var addAction = function(action) {
		if (_curAction)
			_curAction.addAction(action);
		else
			_undoManager.addAction(action);		
	};

	var complete = function() {
		_undoManager.addAction(_curAction);
		_curAction = null;
	};
	
	return {
		init:init,
		addAction:addAction,
		complete:complete
	};
	
}();

function ResetUndoSystem(){
	_aggregateActionMgr.init();
	_undoManager.init();	
}
