
// cbmExport class
var cbmExport = (function () {

    // Private 
	var textFile = null;
	var charIndex = 0;
	var screenIndex = 0;
	var colourIndex = 0;
	var mode = "";
	var inputScreen = 0;
    var _rle = false;
	var _exportBuffer;
	var charImportBuffer = [];
	var singleScreenOnly = false;
	var inputRadix;
	
    // Private methods
    var Load = function (text) {

    };
	
	// Enable starts with
	if (typeof String.prototype.startsWith != 'function') {
	  // see below for better implementation!
	  String.prototype.startsWith = function (str){
		return this.indexOf(str) == 0;
	  };
	}


	var RLEEncode8 = function(sourceArray){

		var outputList = [];
		var count = 1;
		var current = sourceArray[0];

		var rleOutput = function(){
			var addition = 0;
			
			// If we have more than one item or the current item cannot be represented as a single value (> 127)
			if (count > 1 || current >=128 )
			{
				outputList.push(count);	
			}
			else
			{
				addition = 128;
			}
			outputList.push(current + addition);	
		};
		
		var dumpOutput = function () {
			rleOutput();
			count = 1;
			current = loopItem;		
		};

		for (var i = 1; i < sourceArray.length; i++){
			var loopItem = sourceArray[i];
			
			if (count > 126) // Max size exceeded
			{
				dumpOutput();
			}
			else if (current == loopItem) // Tally
			{
				count++;
			}
			else // New char force output
			{
				dumpOutput();
			}
		}

		dumpOutput();
		return outputList;
	};
	
	function ExtractItemValue(value)
	{
		var parseText = value;
		var converted = value;

		try
		{
			
			if (value.toString().startsWith('$'))
			{
				inputRadix=2; 
				parseText = "0x" + value.substring(1);
				converted = parseInt(parseText, 16);
			}
			else if (value.toString().startsWith('0x'))
			{
				inputRadix=1; 
				converted = parseInt(parseText, 16);
			}
			else if (value.toString().startsWith('%'))
			{
				inputRadix=3; 
				parseText = value.substring(1);
				converted = parseInt(parseText, 2);
			}
			else
			{
				inputRadix=0; 
				converted = parseInt(value);
			}	
		}
		catch(err) {
			throw (err);
		}	

		return converted;

	}
		
	function ProcessLineItem(index, value)
	{
		var curScreen;
		
		if (!singleScreenOnly)
		{
			curScreen = CbmScreens[inputScreen];
		}
		else
		{
			curScreen = CbmScreen;
		}
		
		//console.log("Processing data cell :" +  index + "value:" + value);
		var itemValue = ExtractItemValue(value);
		
		switch (mode)
		{
			case "CHAR":
				charImportBuffer[charIndex] = itemValue;
				charIndex++;
				//console.log("Adding char:" + charIndex + "value: " +  itemValue);
				break;
			case "SCREEN":
				if (singleScreenOnly && inputScreen >  0)
					return;
				curScreen.setCellCharFromIndex(screenIndex, itemValue);
				screenIndex++;
				//console.log("Adding cell:" + screenIndex + "value: " +  itemValue);
				break;
			case "COLOUR":
				if (singleScreenOnly && inputScreen >  0)
					return;
				curScreen.setCellColourFromIndex(colourIndex, itemValue);
				colourIndex++;
				//console.log("Adding colour cell:" + colourIndex + "value: " +  itemValue);
				break;			
		}	
	}
	
	function ExtractToken(line, token)
	{
		if (line.startsWith(token))
		{
			var tokLen = token.length;
			var strCol = line.substr(tokLen , 2).trim();
			return strCol;
		}	
		
		return null;
	}
	
	function TrySetControl(line, token, control)
	{
		if (line.startsWith(token))
		{
			var tokVal = ExtractToken(line, token);
			$(control).val(tokVal).trigger('change');				
		}	
	}	

	
	var spareDataItem = null;
	function readRLELine(dataItems)
	{
		var sourceDataItems = dataItems;
		if (spareDataItem != null)
		{
			sourceDataItems.unshift(spareDataItem);
			spareDataItem = null;
		}

		var newDataItems = [];
		while( dataItem = sourceDataItems.shift() ) 
		{
			extractedDataItem = ExtractItemValue(dataItem);
			if (extractedDataItem >=128)
			{
				newDataItems.push(extractedDataItem - 128);
			}
			else
			{
				// If an RLE encoded value not truncated....
				if (dataItems.length >= 1)
				{
					var count = extractedDataItem;
					var itemVal = dataItems.shift();
					var convertedValue = ExtractItemValue(itemVal);
					for (var i = 0; i < count; i++)
						newDataItems.push(itemVal);
				}
				else
				{
					// .. else put it back for next time....
					spareDataItem = dataItem;
				}
			}				  
		}
		
		return newDataItems;
	
	}

	function ProcessLine(line)
	{
		//console.log("Importing line:" + line);

		if (line.startsWith(";rle-encoded=true"))
		{
			$('#chkRLE').prop('checked', true).triggerHandler('change');
			_rle = true;
			return;
		}
		
		if (line.startsWith(";char data"))
		{
			mode = "CHAR";
			//console.log("Mode set to:" + mode);
			return;
		}
		
		if (line.startsWith(";screen data"))
		{
			mode = "SCREEN";
			inputScreen = 0;
			screenIndex = 0;
			if (line.startsWith(";screen data1")) inputScreen = 1;
			if (line.startsWith(";screen data2")) inputScreen = 2;
			if (line.startsWith(";screen data3")) inputScreen = 3;
			if (line.startsWith(";screen data4")) inputScreen = 4;

			//console.log("Mode set to:" + mode);
			return;
		}
		
		if (line.startsWith(";colour data"))
		{
			mode = "COLOUR";
			inputScreen = 0;
			colourIndex = 0;
			if (line.startsWith(";colour data1")) inputScreen = 1;
			if (line.startsWith(";colour data2")) inputScreen = 2;
			if (line.startsWith(";colour data3")) inputScreen = 3;
			if (line.startsWith(";colour data4")) inputScreen = 4;
			//console.log("Mode set to:" + mode);
			return;
		}

		
		// Process geometry changes when not in single mode
		if (!singleScreenOnly)
		{
			TrySetControl(line, ";background-colour=", "#gBackGroundColourSelect");
			TrySetControl(line, ";border-colour=", "#gBorderColourSelect");
			TrySetControl(line, ";aux-colour=", "#gAuxColourSelect");

			TrySetControl(line, ";col-count=", "#screenColumnCount");
			TrySetControl(line, ";row-count=", "#screenRowCount");
			
			if (line.startsWith(";char-height="))
			{
				var tokVal = ExtractToken(line, ";char-height=");
				//alert(tokVal);
				$('#chkDoubleHeight').prop('checked', tokVal != 8).triggerHandler('change');
			}
		}
		
		var bytePos = line.indexOf(".byte");
		if (bytePos >= 0)
		{
			//console.log("Reading data line:" + line);

			// Split into data items
			var dataItems = line.substring(bytePos + 6).trim().split(",");
			
			// Handler RLE
			if (_rle && (mode == "SCREEN" || mode == "COLOUR"))
			{
				dataItems = readRLELine(dataItems);
			}
			
			for (var i =0; i < dataItems.length; i++)
			{
				ProcessLineItem(i, dataItems[i]);
			}
			
			
			
		}
	}

	// Generate export text
	var importFile = function (text) {
		
	 	charImportBuffer = [];
		charIndex = 0;
		screenIndex = 0;
		colourIndex = 0;

		singleScreenOnly = $("#chkSingleScreen").is(':checked');
		
		$('#chkRLE').prop('checked', false).triggerHandler('change');
		_rle = false;

		// Split string into lines
		var lines = text.match(/^.*((\r\n|\n|\r)|$)/gm);
		
		// For each line
		$.each(lines, function( index, value ) {
			ProcessLine(value.trim());
		});
		
		var charWanted = $("#chkLoadChars").is(':checked');
		if (charWanted)
			uDCGen.setFromBuffer(charImportBuffer);
		
		// Show radix read from file..
		$("#outputRadix").val(inputRadix);
		
		primeCharacterTableFromScreen();
		
 	}	

	// Generate export text
	var primeCharacterTableFromScreen = function () {
	
		// Set attributes on chars based on screen usage..
		for (var iScreen = 0; iScreen < 5; iScreen ++)
		{
			var curScreen = CbmScreens[iScreen];
			
			// Force to CBM screen for single screen mode...
			if (singleScreenOnly)
				curScreen = CbmScreen;
			
			// Set colors based on what's in the screen
			for (var i = 0; i < CbmScreen.CharCount; i++)
			{
				var cidx = curScreen.getCharFromIndex(i);
				var charColour = curScreen.getColValFromIndex(i);
				var cbmCharObj = uDCGen.getCbmChar(cidx);
				cbmCharObj.IntendedMultiColor = (charColour >= 8 ) ;
				cbmCharObj.IntendedCharColor = charColour & ~8;
			}
		}
	
 	}		
	
    // Private methods
    var makeTextFile = function (text) {
		var data = new Blob([text], {type: 'text/plain'});

		// If we are replacing a previously generated file we need to
		// manually revoke the object URL to avoid memory leaks.
		if (textFile !== null) {
		  window.URL.revokeObjectURL(textFile);
		}
	  
		textFile = window.URL.createObjectURL(data);
		return textFile;
    };

	function toRadix(num, radix, length) {
		return  ("000000000000000" +(Number((num>>>0)).toString(radix))).slice(-length).toUpperCase()
	}

	function formatOutputValue(value, index)
	{
		
		var fmt = parseInt($(outputRadix).val());
		var retString = "";
		
		switch (fmt)
		{
			case 0: retString = value.toString(); break
			case 1: retString = "0x" + toRadix(value, 16, 2); break
			case 2: retString = "$" + toRadix(value, 16, 2); break
			case 3: retString = "%" + toRadix(value, 2, 8);  break
		}
		
		return retString;

	}

	
		var outputAppendArray = function(sourceArray, rleEncode){

			var outputArray;
			
			if (rleEncode)
			{
				outputArray = RLEEncode8 (sourceArray);
			}
			else
			{
				outputArray = sourceArray;
			}
			
			for (var i = 0; i < outputArray.length; i++)
			{
				groupedAppend8(formatOutputValue(outputArray[i]));
			} 
	};
	
	
	function outputScreen(index, rleWanted) {

		var screen = CbmScreens[index]
		if (!screen.inUse())
			return;

		appendToExport(";screen data"+index);
		appendToExport("");

		// Output the screen array
		var outputArray = screen.getCharBuffer();
		outputAppendArray(outputArray, rleWanted);
		groupedAppend8flush();
		
		groupedAppend8Array = [];
		appendToExport("");
		appendToExport(";colour data"+index);
		appendToExport("");
				
		// Output the screen array
		var outputArray = screen.getColourBuffer();
		outputAppendArray(outputArray, rleWanted);
		groupedAppend8flush();	
		appendToExport("");
	}

	// Generate export text
	var generateExport = function (text) {
	
		var charPrefix = $(charLabelPrefx).val() + "";
		
		 _exportBuffer ="";
		appendToExport(";settings");	
		appendToExport(";background-colour=" + gBackColIdx);
		appendToExport(";border-colour=" + gBorderColIdx);
		appendToExport(";aux-colour=" + gAuxColIdx);
		
		var rleWanted = $("#chkRLE").is(':checked');
		if (rleWanted)
		{
			appendToExport(";rle-encoded=" + rleWanted);
		}

		var rows = $("#screenRowCount").val();
		var cols = $("#screenColumnCount").val();
		var normalSettings = true;
		
		normalSettings &= (gCharHeight == 8);
		normalSettings &= (rows == 23);
		normalSettings &= (cols == 22);
		
		//if (!normalSettings)
		{
			appendToExport(";char-height=" + gCharHeight);
			appendToExport(";row-count=" + rows);
			appendToExport(";col-count=" + cols);
		}
		
		appendToExport("");	
				
		appendToExport(";char data");
		var count = uDCGen.getCharCount();
		for (var i = 0; i < count; i++)
		{
			var charObj = uDCGen.getCbmChar(i);
			var outputValues = $.map( charObj.rows(), formatOutputValue); 
			var usedPrefix = "";
			if (charPrefix.length > 1)
			{
				usedPrefix = charPrefix + i + ": ";
			}
			var output = usedPrefix + ".byte  " + outputValues.join();
			appendToExport(output);
		}

		groupedAppend8Array = [];
		appendToExport("");
				
		// Output all the screens
		for (var i = 0; i < maxScreens; i++)
		{
			outputScreen(i, rleWanted);
		}

		generateCharUsageReport();

		return _exportBuffer;
	}
	



	// Generate export text
	var generateTileMap = function (text) {
	
		_exportBuffer ="";
		outputScreenTileMap(0);
		return _exportBuffer;

		// Output all the screens
		for (var i = 0; i < maxScreens; i++)
		{
			outputScreenTileMap(i);
		}

		return _exportBuffer;
	}

    function outputScreenTileMap(index) {

		var screen = CbmScreens[index]
		if (!screen.inUse())
			return;

		appendToExport(";tile map data "+index);
		appendToExport("");

		// Output the screen array
		var outputArray = screen.getCharBuffer();
		var rows = screen.RowCount;
		var cols = screen.ColCount;

		var pairs = {};

		// Generate a dictionary of unique char pairs...
		for (var r = 0; r < rows; r++){
			for (var c = 0; c < (cols - 1); c++) {
				var c0 = screen.getCharFromRowCol(r,c);
				var c1 = screen.getCharFromRowCol(r, c +1);
				var pair = c0 + ":" + c1;
				pairs[pair] = true;
			}
		}

		//var size = Object.keys(pairs).length; //29
	for(var key in pairs) {
		var value = pairs[key];
			appendToExportAsComment (key);
 
 
}

	}


	function fmt3(x){
		return fmt2(x).replace("00", "--");
	}

	function fmt2(x){
		return toRadix(x, 10, 2);
	}

	function generateCharUsageReport(){
		appendToExport("");
		appendToExportAsComment ("Char Usage Report");
		appendToExportAsComment (" =================                                      00 01 02 03 04");
		var usage = getCharUsage();
		for(var c=0; c<= 255; c++){
			var usObj = usage[c];
			var usText = "";
			for(var screen=0; screen <  maxScreens; screen++) {
				usText+=fmt3(usObj.ScreenInfo[screen].usage) + ",";
			}

			var outText = "Char Index (" + fmt2(c) +") TotalScreenCount:" + fmt2(usObj.ScreenCount) + " ScreenUsage [0-4]: " + usText;
			appendToExportAsComment (outText);
		}

	}
	var groupedAppend8Array = [];

	function groupedAppend8flush()
	{
		appendToExport(".byte  " + groupedAppend8Array.join());
		groupedAppend8Array = [];	
	}

	function groupedAppend8(text)
	{
		if (groupedAppend8Array.length >=8)
		{
			groupedAppend8flush();
		}
		groupedAppend8Array.push(text);
	}
	
	function appendToExportAsComment(text){
		appendToExport ("; "+text );
	}

	function appendToExport(text)
	{
		_exportBuffer += text;
		_exportBuffer += "\r\n";
	}

	
    // Closure Interface
    return {
        makeTextFile: makeTextFile,
		generateExport: generateExport,
		importFile: importFile,
		primeCharacterTableFromScreen : primeCharacterTableFromScreen,
		generateTileMap : generateTileMap
    };

})();
