var maxScreens = 5;
var CbmScreens = [];
var CbmScreen;

// CbmScreen class
var _CbmScreen = (function (rowCount, colCount) {
    
	var blankChar = 32;
	var _colCount = colCount;
	var _rowCount = rowCount;
	var _charCount = _rowCount * _colCount;
	var backupScreenCharRAM = [];
	var screenCharRAM = [];
	var screenColourRAM = [];
	var selectionRect;
	var lastRow = _rowCount - 1;
	var lastCol = _colCount - 1;
	

		// return char by coords
	var setScreenCharAction = function (row,col,charVal, parentAction) {

		var row = row;
		var col = col;
		var charVal = charVal;
		var oldScreenVal = screenCharRAM[row * _colCount + col] ;
			
		if (oldScreenVal == charVal){
			return null;
		}

		var redo = function () {
			screenCharRAM[row * _colCount + col] = charVal;
		};		

		var undo = function () {
			screenCharRAM[row * _colCount + col] = oldScreenVal;
		};	

		return {
			undo:undo,
			redo:redo
			};
		};


		// return char by coords
	var setScreenColourAction = function (row,col,colorVal) {

		var row = row;
		var col = col;
		var colorVal = colorVal;
		var oldColorVal = screenColourRAM[row * _colCount + col] ;
			
		if (oldColorVal == colorVal){
			return null;
		}

		//  return char by index
		var redo = function () {
			screenColourRAM[row * _colCount + col] = colorVal;
		};		

		 //  return char by index
		var undo = function () {
			screenColourRAM[row * _colCount + col] = oldColorVal;
		};	

		return {
			undo:undo,
			redo:redo
			};
		};

		// return char by coords
	var setScreenMC  = function (row,col,mc) {

		var row = row;
		var col = col;
		var mc = mc;
		var oldMC = getCellMultiColourMode(row,col);
		
		if (oldMC == mc){
			return null;
		}

		//  return char by index
		var redo = function () {
			doSetCellMultiColourMode(row,col,mc);
		};		

		 //  return char by index
		var undo = function () {
			doSetCellMultiColourMode(row,col,oldMC);
		};	

		return {
			undo:undo,
			redo:redo
			};
		};

		
		//  return char by index
    var getCharBuffer = function () {
        return screenCharRAM;
    };

     //  return char by index
    var getColourBuffer = function () {
        return screenColourRAM;
    };
	
	// Declare and allocate an array to hold the UDC data
    var cbmCharVals = [];
    for (var i = 0; i < _charCount; i++)
	{
        screenCharRAM.push(blankChar);
        screenColourRAM.push(0);
	}

     //  return char by index
    var screenHeight = function () {
        return _rowCount * gCharHeight;
    };
	
     //  return char by index
    var screenWidth = function () {
        return _colCount * 8;
    };

    // return char by coords
    var getCoOrdsAt = function (row,col) {
		checkRange(row,col);
        var x = col * 8;
        var y = row * gCharHeight;
		return {"x":x,"y":y};
    };

	//  return char by index
    var getCharFromIndex = function (index) {
		checkIndex(index);
        return screenCharRAM[index];
    };

	//  return char by index
    var getCharUsageCount = function (charVal) {
		var count = 0;
		for (var i =0; i < _charCount; i++)
		{
			if (screenCharRAM[i] == charVal)
				count++;
		}
		return count;
    };	

	//  return char by index
    var prepareRelocate = function () {
		backupScreenCharRAM = screenCharRAM.slice(0);
    };	

	//  return char by index
    var relocate = function (curIndex, newIndex) {
		//alert("relocate " +curIndex +" " + newIndex);
		var count = 0;
		for (var i =0; i < _charCount; i++)
		{
			if (backupScreenCharRAM[i] == curIndex)
			{
				screenCharRAM[i] = newIndex;
			}
		}
		return false;
    };	

	//  return char by index
    var inUse = function () {
		var count = 0;
		for (var i =0; i < _charCount; i++)
		{
			if (screenCharRAM[i] != 32)
			{
				return true;
			}
		}
		return false;
    };	
	
	//  return char by index
    var getColValFromIndex = function (index) {
		checkIndex(index);
        return screenColourRAM[index];
    };
	
    // return char by coords
    var getCharFromRowCol = function (row,col) {
		checkRange(row,col);
        return screenCharRAM[row * _colCount + col];
    };

    // return char by coords
    var getCharColourFromRowCol = function (row,col) {
		checkRange(row,col);
        return screenColourRAM[row * _colCount + col];
    };
	
    // return char by coords
    var setChar = function (row,col,charVal,colorVal) {
		checkRange(row,col);
		
		// Set the multi colour bit if global flag set...
		if (gIntendedMultiColour)
			colorVal|= 8;
        
		screenCharRAM[row * _colCount + col] = charVal;
		screenColourRAM[row * _colCount + col] = colorVal;
    };
	
    // return char by coords
    var setCellChar = function (row,col, charVal) {
		checkRange(row,col);
		
		var action = setScreenCharAction(row,col, charVal);
		if (action != null)
		{
			action.redo();
			_aggregateActionMgr.addAction(action);
		}
		//screenCharRAM[row * _colCount + col] = charVal;
    };
	
    // return char by coords
    var setCellColour = function (row,col, colorVal, ignoreMC) {
		checkRange(row,col);
		var colourToSet;
		if (ignoreMC != null && ignoreMC != undefined)
		{
			var curVal = screenColourRAM[row * _colCount + col];
			var curMC = curVal & 8;
			var newColur = (colorVal & ~8) | curMC;
			colourToSet = newColur;
		}		
		else
		{
			colourToSet = colorVal;
		}
		
		//screenColourRAM[row * _colCount + col] = colourToSet;		
		var action = setScreenColourAction(row,col, colourToSet);
		if (action != null)
		{
			action.redo();
			_aggregateActionMgr.addAction(action);
		}
		
    };

    // return char by coords
    var setCellMultiColourMode = function (row,col, multiColourMode) {
		checkRange(row,col);
		var action = setScreenMC(row,col, multiColourMode);
		if (action != null)
		{
			action.redo();
			_aggregateActionMgr.addAction(action);
		}
    };

	// return char by coords
    var doSetCellMultiColourMode = function (row,col, multiColourMode) {
		checkRange(row,col);
		if (multiColourMode)
			screenColourRAM[row * _colCount + col] |= 8;
		else
			screenColourRAM[row * _colCount + col] &= ~8;
    };

    // return char by coords
    var getCellMultiColourMode = function (row,col) {
		checkRange(row,col);
		return (screenColourRAM[row * _colCount + col] & 8) == 8;
    };
	
    var clearSelectionRect = function () {
		this.selectionRect = null;
    };	
	
	var selectionWidthPixels = function () {
		if (this.selectionRect)
			return this.selectionRect.width * 8;
		else
			return null;
    };	

	var selectionHeightPixels = function () {
		if (this.selectionRect)
			return this.selectionRect.height * gCharHeight;
  		else
			return null;  
	};	

	
	var selectAll = function () {
		this.selectionRect = {row:0, col:0, width: _colCount-1, height: _rowCount-1};
    };

	var keyDown = function (charVal, color) {
		if (this.selectionRect != null && this.selectionRect.row>=0 && this.selectionRect.col>=0)
		{
			if (this.selectionRect.width < 1)
				if (this.selectionRect.height < 1)
				{
					_aggregateActionMgr.init();	
					setCellChar(this.selectionRect.row, this.selectionRect.col, charVal);
					setCellColour(this.selectionRect.row, this.selectionRect.col, color);
					_aggregateActionMgr.complete();	
					cursorNext.bind(this)();
				}
		}
    };

	var cursorPrev = function () {
		if (this.selectionRect.col == 0 && this.selectionRect.row > 0)
		{
			offsetSelectionRect.bind(this)((_colCount-1),-1);
		}
		else if (this.selectionRect.col == 0 && this.selectionRect.row == 0)
		{
		}
		else
		{
			offsetSelectionRect.bind(this)(-1,0);
		}	
	}
	
	var cursorNext = function () {

		if (this.selectionRect.col >= lastCol && this.selectionRect.row < lastRow)
		{
			offsetSelectionRect.bind(this)(-(_colCount-1),1);
		}
		else if (this.selectionRect.col == lastCol && this.selectionRect.row == lastRow)
		{
		}
		else
		{
			offsetSelectionRect.bind(this)(1,0);
		}	
	}
	
	var newLine = function (charVal, color) {
		if (this.selectionRect != null && 
				this.selectionRect.row>=0 && 
				this.selectionRect.row < _rowCount - 1 && 
				this.selectionRect.col>=0
				)
		{
			if (this.selectionRect.width < 1)
				if (this.selectionRect.height < 1)
				{
					offsetSelectionRect.bind(this)(-_colCount,1);
					this.selectionRect.col = 0;
				}
		}
    };	
	
	// return char by coords
    var setSelectionRect = function (row,col,width,height) {
		this.selectionRect = {row:row, col:col, width: width, height:height};
    };

    // return char by coords
    var setCellCharFromIndex = function (index, charVal) {
		checkIndex(index);
		screenCharRAM[index] = charVal;
    };
	
    // return char by coords
    var setCellColourFromIndex = function (index, colorVal) {
		checkIndex(index);
		screenColourRAM[index] = colorVal;
    };
	
    // return char by coords
    var getSelectionRect = function () {
		
		if (this.selectionRect != null && this.selectionRect != undefined)
			return {row:this.selectionRect.row, col:this.selectionRect.col, width: this.selectionRect.width, height:this.selectionRect.height};
		return this.selectionRect;
    };
	
    // return char by coords
    var offsetSelectionRect = function (cx,cy) {
		
		if (this.selectionRect != null && this.selectionRect != undefined)
		{
			var newRow = this.selectionRect.row + cy;
			var newCol = this.selectionRect.col + cx;
			
			if (newRow <= lastRow && newRow >=0)
				this.selectionRect.row = newRow; 
			if (newCol <= lastCol && newCol >=0)
				this.selectionRect.col = newCol;
		}
		return this.selectionRect;
    };
	
    // return char by coords
    var inflateSelectionRect = function (cx,cy) {
		
		if (this.selectionRect != null && this.selectionRect != undefined)
		{
			this.selectionRect.width+=cx; 
			this.selectionRect.height+=cy;
		}
		return this.selectionRect;
    };


	
    // row col index range check
	var checkRange = function (row, col) {
        if ((row < 0) || (row > _rowCount))
            throw "row out of range";
        if ((col < 0) || (col > _colCount))
            throw "col out of range";
    };

	// Screen index range check
	var checkIndex = function (idx) {
	if ((idx < 0) || (idx > _charCount))
		throw "row out of range";
	};
	
    // Closure Interface
    return {
		getCharFromIndex : getCharFromIndex,
		getColValFromIndex :getColValFromIndex,
		getCharFromRowCol : getCharFromRowCol,
		ColCount : _colCount,
		RowCount : _rowCount,
		CharCount : _charCount,
		getCoOrdsAt : getCoOrdsAt,
		screenHeight : screenHeight,
		screenWidth : screenWidth,
		setChar : setChar,
		getSelectionRect : getSelectionRect,
		setSelectionRect : setSelectionRect,
		getCharColourFromRowCol : getCharColourFromRowCol,
		setCellChar :setCellChar,
		setCellColour :setCellColour,
		setCellMultiColourMode : setCellMultiColourMode,
		getCellMultiColourMode : getCellMultiColourMode,
		getCharUsageCount: getCharUsageCount,
		setCellCharFromIndex : setCellCharFromIndex,
		setCellColourFromIndex: setCellColourFromIndex,
		clearSelectionRect: clearSelectionRect,
		getCharBuffer: getCharBuffer,
		getColourBuffer : getColourBuffer,
		offsetSelectionRect : offsetSelectionRect,
		inflateSelectionRect : inflateSelectionRect,
		selectAll : selectAll,
		inUse: inUse,
		keyDown : keyDown,
		newLine : newLine,
		cursorPrev : cursorPrev,
		cursorNext : cursorNext,
		relocate : relocate,
		prepareRelocate : prepareRelocate,
		selectionWidthPixels : selectionWidthPixels,
		selectionHeightPixels : selectionHeightPixels
    };
});

var _CbmScreenArray = (function (curIndex, newIndex) {

    var prepareRelocate = function () {
		for (var i = 0; i < maxScreens; i++)
		{
			CbmScreens[i].prepareRelocate();
		}
    };	
	
    var relocate = function (curIndex, newIndex) {
		for (var i = 0; i < maxScreens; i++)
		{
			CbmScreens[i].relocate(curIndex, newIndex);
		}
    };	

	//  return char by index
    var getCharUsageCount = function (index) {
		var count = 0;
		for (var i = 0; i < maxScreens; i++)
		{
			count += CbmScreens[i].getCharUsageCount(index);
		}
		return count;
    };	
	
    return {
		getCharUsageCount : getCharUsageCount,
		relocate : relocate,
		prepareRelocate : prepareRelocate
    };
})();

function createScreenObject(rows, cols)
{
	for (var i = 0; i < maxScreens; i++)
	{
		CbmScreens[i] = new _CbmScreen(rows, cols);
	}
	CbmScreen = CbmScreens[0];
}

function showSelectedScreen()
{
	var screenNum = parseInt($("#screenNum").val());
	CbmScreen = CbmScreens[screenNum];
	$('#charUsageCount').text(CbmScreen.getCharUsageCount(gCurrentCanvasChar));
}

function  createScreens()
{
	var rows = $("#screenRowCount").val();
	var cols = $("#screenColumnCount").val();
	createScreenObject(rows,cols)
}

createScreens();
showSelectedScreen();
