
// cbmExport class
var miniPaintIO = (function () {

	var MINIPAINT_BASECHAR = 16;
	var LOAD_ADDRESS_OFFSET = 0;
	var LOAD_ADDRESS_SIZE   = 2;
	var BASICSTUB_OFFSET = LOAD_ADDRESS_OFFSET + LOAD_ADDRESS_SIZE;
	var BASICSTUB_SIZE   = 13;
	var VICREG_OFFSET = BASICSTUB_OFFSET + BASICSTUB_SIZE;
	var VICREG_SIZE   = 2;
	var BITMAP_OFFSET = VICREG_OFFSET + VICREG_SIZE;
	var BITMAP_SIZE   = 3840;
	var COLRAM_OFFSET = BITMAP_OFFSET + BITMAP_SIZE;
	var COLRAM_SIZE   = 120;
	var DISPCODE_OFFSET = COLRAM_OFFSET + COLRAM_SIZE;
	var DISPCODE_SIZE   = 120;

	String.prototype.b = function() { return parseInt(this,2); }

	var auxRegMASK  = "11110000".b();
	var backColMASK = "11110000".b();
	var borColMASK =  "00000111".b();
	var reader;

	var saveByteArray = (function () {
		var a = document.createElement("a");
		document.body.appendChild(a);
		a.style = "display: none";
		return function (data, name) {
			var blob = new Blob(data, {type: "octet/stream"}),
				url = window.URL.createObjectURL(blob);
			a.href = url;
			a.download = name;
			a.click().trigger();
			window.URL.revokeObjectURL(url);
		};
	}());

	function initMiniPaintMode()
	{
		ResetUndoSystem();
		
		// Set screen size to 12 x 20 DH
		$("#screenColumnCount").val(20).trigger('change');
		$("#screenRowCount").val(12).trigger('change');
		$("#chkDoubleHeight").prop('checked', true).triggerHandler('change');	

		// Push graphic data into characters
		var countOfChars = BITMAP_SIZE / gCharHeight;
		
		// Array of zeros
		var charData = Array.apply(null, new Array(gCharHeight)).map(Number.prototype.valueOf,0);

		
		for (var count = MINIPAINT_BASECHAR; count < MINIPAINT_BASECHAR + countOfChars; count++)
		{
			var bufIndex = BITMAP_OFFSET + (gCharHeight * count);
			var targetChar = uDCGen.getCbmChar(count);
			targetChar.setData(charData);
		}
		
		// Poke chars to screen in column major order
		var charIdx = MINIPAINT_BASECHAR;
		for (var col = 0; col < 20; col++)
		{
			for (var row = 0; row < 12; row++)
			{
				CbmScreen.setCellChar(row, col, charIdx++);
			}
		}	

		
		
		RefreshModifiedTableImages();
		RedrawScreenAndCanvas();
		SetCurCharFocus(gCurrentCanvasChar);
		ResetUndoSystem();		
	}
	
	
	function Export()
	{
		var formatOK = true;
		formatOK &= $("#screenColumnCount").val() == 20;
		formatOK &= $("#screenRowCount").val() == 12;
		formatOK &= $("#chkDoubleHeight").is(':checked');
		
		if (!formatOK)
		{
			alert ("Drawing is not compatible with miniPaint. Requires [12 rows, 20 cols, 16x8chars). Note: Adjusting now will LOSE YOUR CURRENT WORK");
			return;
		}



		
		var loaderData = [0xF1, 0x10, 0x0C, 0x12, 0xD8, 0x07, 0x9E, 0x20, 0x38, 0x35, 0x38, 0x34, 0x00, 0x00, 0x00];
		var dispRoutine = [0x18, 0xA9, 0x10, 0xA8, 0x99, 0xF0, 0x0F, 0x69, 0x0C, 0x90, 0x02, 0xE9, 0xEF, 0xC8, 0xD0, 
							0xF4, 0xA0, 0x05, 0x18, 0xB9, 0xE4, 0xED, 0x79, 0xFA, 0x21, 0x99, 0x00, 0x90, 0x88, 0x10, 
							0xF3, 0xAD, 0x0E, 0x90, 0x29, 0x0F, 0x0D, 0x0E, 0x12, 0x8D, 0x0E, 0x90, 0xAD, 0x0F, 0x12, 
							0x8D, 0x0F, 0x90, 0xA9, 0x10, 0x85, 0xFB, 0xA9, 0x12, 0x85, 0xFC, 0xA9, 0x00, 0x85, 0xFD, 
							0xA9, 0x11, 0x85, 0xFE, 0xA2, 0x0F, 0xA0, 0x00, 0xB1, 0xFB, 0x91, 0xFD, 0xC8, 0xD0, 0xF9, 
							0xE6, 0xFC, 0xE6, 0xFE, 0xCA, 0xD0, 0xF2, 0xA2, 0x00, 0xA0, 0x00, 0xBD, 0x10, 0x21, 0xE8, 
							0x99, 0x00, 0x94, 0xC8, 0x4A, 0x4A, 0x4A, 0x4A, 0x99, 0x00, 0x94, 0xC8, 0xC0, 0xF0, 0xD0, 
							0xEC, 0x20, 0xE4, 0xFF, 0xF0, 0xFB, 0x4C, 0x32, 0xFD, 0x02, 0xFE, 0xFE, 0xEB, 0x00, 0x0C];
	

	var outputbuffer = new ArrayBuffer(4097);
		var outputArray = new Int8Array(outputbuffer);		

		// Copy loader to output buffer
		for (var i = 0; i < LOAD_ADDRESS_SIZE + BASICSTUB_SIZE; i++)
			outputArray[LOAD_ADDRESS_OFFSET+i] = loaderData[i];
		
		// Copy disp routine to output buffer
		for (var i = 0; i < DISPCODE_SIZE; i++)
			outputArray[DISPCODE_OFFSET+i] = dispRoutine[i];		

		// Store VICREG 
		outputArray[VICREG_OFFSET] = gAuxColIdx << 4;
		outputArray[VICREG_OFFSET+1] = (gBackColIdx << 4) | gBorderColIdx | 8;
		
		// Store char ram
		var charIdxLineIdx = 0;
		for (var col = 0; col < 20; col++)
		{
			for (var row = 0; row < 12; row++)
			{
				// Get the character at for the screen location
				var charIndex = CbmScreen.getCharFromRowCol(row, col);
				var cbmCharObj = uDCGen.getCbmChar(charIndex);
				
				for (var charRow=0; charRow < gCharHeight; charRow++)
				{
					outputArray[BITMAP_OFFSET + charIdxLineIdx] = cbmCharObj.row(charRow);
					charIdxLineIdx++;
				}
			}
		}

		// Store colour ram
		var screenColIndex = 0;
		var outputColIndex = 0;
		for (var i = 0;  i < (12 * 10); i++)
		{
			var lowerNibble = CbmScreen.getColValFromIndex(screenColIndex++);
			var upperNibble = CbmScreen.getColValFromIndex(screenColIndex++);
			var mpColourVal = (upperNibble << 4) | lowerNibble
			outputArray[COLRAM_OFFSET + outputColIndex] = mpColourVal;
			outputColIndex++;
		}	

		saveByteArray([outputArray], 'exportMP.prg');
		
	}
	
	function ProcessImport(e)
	{

		// Read as an array buffer
		var rawData = reader.result;
	
		// obtain a byte view
		var mPaintBuf = new Int8Array(rawData);		
		
		// Decode VICREG and extract colours
		var auxColor = (mPaintBuf[VICREG_OFFSET] & auxRegMASK) >> 4;	
		var borderCol = (mPaintBuf[VICREG_OFFSET+1] & borColMASK);	
		var backCol =   (mPaintBuf[VICREG_OFFSET+1] & backColMASK) >> 4;		
		
		$("#gAuxColourSelect").val(auxColor).trigger('change');
		$("#gBorderColourSelect").val(borderCol).trigger('change');
		$("#gBackGroundColourSelect").val(backCol).trigger('change');

		// Set screen size to 12 x 20 DH
		$("#screenColumnCount").val(20).trigger('change');
		$("#screenRowCount").val(12).trigger('change');
		$("#chkDoubleHeight").prop('checked', true).triggerHandler('change');

		// Push graphic data into characters
		var countOfChars = BITMAP_SIZE / gCharHeight;
		
		for (var count = 0; count < countOfChars; count++)
		{
			var bufIndex = BITMAP_OFFSET + (gCharHeight * count);
			var targetChar = uDCGen.getCbmChar(MINIPAINT_BASECHAR + count);
			var charData = mPaintBuf.subarray(bufIndex, bufIndex + gCharHeight);
			targetChar.setData(charData);
		}
		
		// Poke chars to screen in column major order
		var charIdx = MINIPAINT_BASECHAR;
		for (var col = 0; col < 20; col++)
		{
			for (var row = 0; row < 12; row++)
			{
				CbmScreen.setCellChar(row, col, charIdx++);
			}
		}	

		var miniPaintColourIndex = COLRAM_OFFSET;
		var colorRamIndex = 0;
		var upperMask = "11110000".b();
		var lowerMask = "00001111".b();
		// Poke color in row major order
		for (var i = 0;  i < (12 * 10); i++)
		{
			var colorVal = mPaintBuf[miniPaintColourIndex++];
			var upperNibble = (colorVal &  upperMask) >> 4;
			var lowerNibble = colorVal & lowerMask;
			CbmScreen.setCellColourFromIndex(colorRamIndex++, lowerNibble);
			CbmScreen.setCellColourFromIndex(colorRamIndex++, upperNibble);
		}			
		
		cbmExport.primeCharacterTableFromScreen();
		
		RefreshModifiedTableImages();
		RedrawScreenAndCanvas();
		SetCurCharFocus(gCurrentCanvasChar);
	}	
		
	function Import(file)
	{
		reader = new FileReader();
		reader.onload = ProcessImport;
		reader.readAsArrayBuffer(file);
	}	

   return {
        Import: Import,
		Export :Export,
		initMiniPaintMode : initMiniPaintMode
		};

})();
