var mouseDownPos = null;
var screenCopyBuffer = null;

function normaliseRect(rect)
{
	if (rect != null && rect != undefined)
	{
		if (rect.height < 0)
		{
			rect.row += rect.height;
			rect.height *= -1; 
		}
		if (rect.width < 0)
		{
			rect.col += rect.width;
			rect.width *= -1; 
		}
	}
	return rect;
}

function drawScreen(canvas, hintCol, hintRow, hintupdateChar, hintRect) {
	// Refresh char count for focused char

	var context = canvas.getContext("2d");
	prepareCanvas(canvas, context, CbmScreen.screenWidth(), CbmScreen.screenHeight());
    
	var drawMode = parseInt($('#outputMode').val());
	
	// If not drawing a single character...
	if (typeof hintRow === "undefined" && typeof hintCol === "undefined" && typeof hintRect === "undefined")
	{
		context.clearRect(0, 0, canvas.width, canvas.height);
	}

	// Take a copy of global fill colours
	 var fillStyles = gmcFillStyles.slice(0); 
	
	var startRow = 0;
	var startcol = 0;
	var endRow = CbmScreen.RowCount;
	var endCol = CbmScreen.ColCount;
	var selectionRect = CbmScreen.getSelectionRect();
	if (selectionRect != null && selectionRect != undefined)
		normaliseRect(selectionRect);

	if (hintRect != null && hintRect != undefined)
		normaliseRect(hintRect);
		
	for (var row = startRow; row < endRow; row++){

		// If a column has been specified reject other columns
		if (typeof hintRect !== "undefined")
		{
			if (row < hintRect.row)
				continue;
			if (row > hintRect.row+hintRect.height)
				continue;
		}

	// If a column has been specified reject other columns
		if (typeof hintRow !== "undefined" && hintRow >= 0)
		{
			if (row !== hintRow)
				continue;
		}
			
		for (var col = startcol; col < endCol; col++) {
			
			// If a column has been specified reject other columns
			if (typeof hintRect !== "undefined")
			{
				if (col < hintRect.col)
					continue;
				if (col > hintRect.col+hintRect.width)
					continue;
			}
			// If a column has been specified reject other columns
			if (typeof hintCol !== "undefined" && hintCol >= 0)
			{
				if (col !== hintCol)
					continue;
			}
			
			var pos = CbmScreen.getCoOrdsAt(row,col);
			var screenChar = CbmScreen.getCharFromRowCol(row,col);

			// If a column has been specified reject other columns
			if (typeof hintupdateChar !== "undefined")
			{
				if (screenChar != hintupdateChar)
					continue;
			}
		
			var sourceScreenColorIndex = CbmScreen.getCharColourFromRowCol(row, col);
			var screenColorIndex = sourceScreenColorIndex;
			
			// Work out if supposed to be multicolour
			var mc = screenColorIndex & 8;
			
			// Mask out multi colour bit for select first 8 characters
			screenColorIndex = screenColorIndex & ~8;
			
			// Pick correct char colour from  drop down based on screen colour
			var colorStyle = gCharColours[screenColorIndex];
			
			// Store in styles array
			fillStyles[2] = [colorStyle];

			//console.log("drawing at:" + pos.x  + "," + pos.y + " char= " + screenChar);
			context.clearRect(pos.x, pos.y, 8, gCharHeight);
			switch (drawMode)
			{
				case 0:
					var cbmChar = uDCGen.getCbmChar(screenChar);
					blitCachedCharBits(mc, false, fillStyles, canvas, context, cbmChar, pos.x, pos.y, 8, gCharHeight, sourceScreenColorIndex);
					break;
				case 1:
					context.fillStyle = "black";
					context.fillRect(pos.x, pos.y, 8, gCharHeight);
					if (mc)
					{
						context.fillStyle = "white";
						context.strokeStyle = "white";
						context.fillRect(pos.x + 2, pos.y + 2, 4,4);
					}
					break;
				case 2:
					context.fillStyle = fillStyles[2];
					context.fillRect(pos.x, pos.y, 8,gCharHeight);
					break;
			}
			
			if (gScreenGrid)
			{
				context.strokeStyle="grey";
				context.lineWidth=.1;
				context.strokeRect(pos.x, pos.y, 8, gCharHeight);
			}
			
			// If to display a selected cell
			var selectedCell = (gCursorMode === 0);
			if (selectedCell)
			{
				if (selectionRect != null && selectionRect != undefined)
				{
					selectedCell &= (col >= selectionRect.col); 
					selectedCell &= (col <= selectionRect.col + selectionRect.width); 
					selectedCell &= (row >= selectionRect.row); 
					selectedCell &= (row <= selectionRect.row + selectionRect.height); 
				}

				if (selectedCell)
				{
					context.strokeStyle="blue";
					if (selectionRect && initalPos && col == initalPos.col && row == initalPos.row && 
						(selectionRect.width > 1 ||  selectionRect.height > 1))
					{
						context.lineWidth=1.5;
					}
					else
					{
						context.lineWidth=.5;
					}
					
					context.strokeRect(pos.x, pos.y, 8, gCharHeight);
				}			
			}
		}
	}
}

function blitCachedCharBits(multiColour, grid, fillStyles, canvas, context, cbmChar, x, y, width, height, screenColorIndex) {
	
	var bmpCanvas = cbmChar.Canvas();

	if (bmpCanvas.width < 1 || (!(bmpCanvas.screenColorIndex)) || bmpCanvas.screenColorIndex != screenColorIndex){
		bmpCanvas.width = width;
		bmpCanvas.height = height;
		var bmCtx = bmpCanvas.getContext("2d");
		bmpCanvas.style.imageRendering = 'pixelated';
		bmpCanvas.screenColorIndex = screenColorIndex;
		blitCharBits(multiColour, grid, fillStyles, bmpCanvas, bmCtx, cbmChar, 0, 0, bmpCanvas.width, bmpCanvas.height);
	}
	context.imageSmoothingEnabled = false;
    context.drawImage(bmpCanvas, x, y);
}


function getMouseCharPos(e) {


	var curCanvasWidth = $('#screenCanvas').width();
	var curCanvasHeight = $('#screenCanvas').height();

	var xScale =  originalCanvasWidth/curCanvasWidth;
	var yScale =  originalCanvasHeight/curCanvasHeight;
	
	var canvas = e.target;
	var coords = getCursorPosition(e);
	var col = Math.floor(coords.x / canvas.width * xScale * CbmScreen.ColCount);
	var row = Math.floor(coords.y / canvas.height * yScale * CbmScreen.RowCount);
	return {row:row, col:col};
}

function proceessMouseMove(e) {

	var pos = getMouseCharPos(e);

	if (mouseCaptured )
	{
		var curColour  = $("#selectIntendedCharColour").val();
		var paintChar = mouseDownEvent.which === 3 ? 32 : gCurrentCanvasChar;
		
		var width = pos.col - mouseDownPos.col;
		var height = pos.row - mouseDownPos.row;
		
		//alert(gCursorMode);
		switch(gCursorMode)
		{
			case 0:
				CbmScreen.setSelectionRect(mouseDownPos.row,mouseDownPos.col,width,height);
				//console.log (pos.row,pos.col,width,height);
				drawScreen($("#screenCanvas")[0]);	
				var cusText = "" + mouseDownPos.row + "," + mouseDownPos.col + " (" + (width+1) + ":" + (height+1) + ")";
				$("#cursorPos").text(cusText);
				break;
			case 1:

			
				if (e.ctrlKey) {
					if ((pos.col != initalPos.col) && (pos.row != initalPos.row))
						break;
				}
 
			
				//CbmScreen.setChar(pos.row, pos.col, paintChar, curColour);

				var applyChar = $("#chkApplyChar").is(':checked');
				var applyColour = $("#chkApplyCol").is(':checked');
				var applyMC = $("#chkApplyMC").is(':checked');
				var mcWanted = $("#chkMultiColour").is(':checked');
				var colr = curColour & ~8;

				if (applyChar)
				{
					CbmScreen.setCellChar(pos.row, pos.col, paintChar);
				}
				if (applyColour)
				{
					CbmScreen.setCellColour(pos.row, pos.col, colr, true);
				}
				if (applyMC)
				{
					CbmScreen.setCellMultiColourMode(pos.row, pos.col, mcWanted);
				}

				drawScreen($("#screenCanvas")[0], pos.col, pos.row);	
				break;
			default:
				break;
		}

		//alert("Setting location row: " + row + ", col:" + col + " to: " + gCurrentCanvasChar);
	}
	else
	{
		$("#cursorPos").text("" + pos.row + " , " +  pos.col);
	}
}


var initalPos;


$("#screenCanvas").mousedown(function(e) {
	//console.log("mousedown");
	mouseDownPos = getMouseCharPos(e);
	initalPos = mouseDownPos;
	
	if (gCursorMode ===0)
	{
		// Show selected character
		var screenChar = CbmScreen.getCharFromRowCol(mouseDownPos.row, mouseDownPos.col);	
		// Don't set to space...
		if (screenChar != 32)
		{
			var col = CbmScreen.getCharColourFromRowCol(mouseDownPos.row, mouseDownPos.col);
			var mc = (col & 8) == 8;
			var cv = (col & ~8);
			SetCurCharFocusFromScreen(screenChar, mc, cv);	
		}
	}
	else
	{
		// Draw mode
		// Record all actions until mouse up as a sinngle group
		_aggregateActionMgr.init();	
	}
	
	mouseCaptured = true;
	mouseDownEvent = e;
});

$("#screenCanvas").mouseleave(function(e) {
	//console.log("mouseleave");
	$("#cursorPos").text("");

	if (mouseCaptured)
	{
		if (gCursorMode > 0)
			_aggregateActionMgr.complete();	
	}
	mouseCaptured = false;
}); 

$("#screenCanvas").mouseup(function(e) {
	//console.log("mouseup");
	if (mouseCaptured)
	{
		proceessMouseMove(e);
		if (gCursorMode > 0)
			_aggregateActionMgr.complete();	

	}
	
	// Refresh char count for focused char
	$('#charUsageCount').text(CbmScreen.getCharUsageCount(gCurrentCanvasChar));
	RefreshTableCharCount();

	mouseCaptured = false;
});  

$("#screenCanvas").mousemove(function(e) {
	//console.log("mousemove");
   
	proceessMouseMove(e);

});

// Disable context menu
$('#screenCanvas').bind('contextmenu', function(e){
    return false;
}); 


function fillSelection(charValue, transpose){
	var selRect = CbmScreen.getSelectionRect();
	var applyColour = $("#chkApplyCol").is(':checked');
	var applyMC = $("#chkApplyMC").is(':checked');

	if (selRect != null && selRect != undefined)
	{
		normaliseRect(selRect);
		var startRow = selRect.row;
		var startcol = selRect.col;
		var endRow = startRow + selRect.height;
		var endCol = startcol + selRect.width;
		
		_aggregateActionMgr.init();	

		if (transpose != undefined && transpose)
		{
			for (var col = startcol; col <= endCol; col++)
				for (var row = startRow; row <= endRow; row++)
					fillSelectionHelper();
		}
		else
		{
			for (var row = startRow; row <= endRow; row++)
					for (var col = startcol; col <= endCol; col++)
						fillSelectionHelper();
		}

		_aggregateActionMgr.complete();	
		
		RedrawScreen();			
	}

	function fillSelectionHelper(){	

		if (transpose != undefined)
		{
			charValue++;
			// Wrap arround
			if (charValue > 255)
				charValue = 0;
		}

		if (charValue != null)
		{
			CbmScreen.setCellChar(row, col, charValue);
		}
		if (applyColour)
		{
			CbmScreen.setCellColour(row, col, gIntendedCharColSelection, true);
		}
		if (applyMC)
		{
			CbmScreen.setCellMultiColourMode(row, col, gIntendedMultiColour);
		}	
	}
}



// Clear Selection
$("#clearSelection").click(function(e) {
	fillSelection(32);
}); 

// Clear Selection
$("#btnSelectAll").click(function(e) {
	$('#btcModeSel').trigger('click');
	CbmScreen.selectAll();
	RedrawScreen();
}); 

$("#setIntendedChar").click(function(e) {
	var applyChar = $("#chkApplyChar").is(':checked');
	if (applyChar)
		fillSelection(gCurrentCanvasChar);
	else
		fillSelection(null);
	
}); 

// FlipVert
$("#copySelection").click(function(e) {

	var selRect = CbmScreen.getSelectionRect();
	
	if (selRect != null && selRect != undefined)
	{
		screenCopyBuffer = [];
		normaliseRect(selRect);
		var startRow = selRect.row;
		var startcol = selRect.col;
		var endRow = startRow + selRect.height;
		var endCol = startcol + selRect.width;
		
		var offsetRow=0;
		var offsetCol=0;
		for (var row = startRow; row <= endRow; row++){
				for (var col = startcol; col <= endCol; col++) {
					var chr = CbmScreen.getCharFromRowCol(row, col);
					var clr = CbmScreen.getCharColourFromRowCol(row, col);
					var copiedChar = {row:offsetRow, col:offsetCol, chr: chr, clr: clr};
					screenCopyBuffer.push(copiedChar);
					offsetCol++;
				}
				offsetRow++;
				offsetCol=0;
				
			}
	}
}); 

// FlipVert
$("#pasteSelection").click(function(e) {

	if (screenCopyBuffer == null)
		return;


	var applyChar = $("#chkApplyChar").is(':checked');
	var applyColour = $("#chkApplyCol").is(':checked');
	var applyMC = $("#chkApplyMC").is(':checked');
	
	var selRect = CbmScreen.getSelectionRect();
	var startRow = selRect.row;
	var startCol = selRect.col;	

	_aggregateActionMgr.init();	
	for (var i=0; i < screenCopyBuffer.length; i++)
	{
		var elem = screenCopyBuffer[i];
		var targetRow = startRow + elem.row;
		var targetCol = startCol + elem.col;
		
		if (targetRow >= CbmScreen.RowCount)
			continue;

		if (targetCol >= CbmScreen.ColCount)
			continue;
		
		//CbmScreen.setChar(targetRow, targetCol, elem.chr, elem.clr);

		
		var mc = elem.clr & 8;
		var colr = elem.clr & ~8;
		
		if (applyChar)
		{
			CbmScreen.setCellChar(targetRow, targetCol, elem.chr);
		}
		if (applyColour)
		{
			CbmScreen.setCellColour(targetRow, targetCol, colr, true);
		}
		if (applyMC)
		{
			CbmScreen.setCellMultiColourMode(targetRow, targetCol, mc);
		}
	}
	_aggregateActionMgr.complete();	

	RedrawScreen();	
}); 

$("#btnUndo").click(function(e) {

	_undoManager.undoLast();
	RedrawScreen();	
}); 


$("#btnRedo").click(function(e) {

	_undoManager.redoLast();
	RedrawScreen();	
}); 

$("#btnEditBits").click(function(e) {
	$("#dlgBitmap").dialog({title: "Bitmap Editor"});
	_BitMapEditor.onShown();
}); 

$("#btnPrepareRows").click(function(e) {
	fillSelection(gCurrentCanvasChar -1, false);
}); 

$("#btnPrepareCols").click(function(e) {
	fillSelection(gCurrentCanvasChar -1, true);
}); 

